@@ -1,6 +1,6 @@
 import vectorian.core as core
 
-from tqdm import tqdm
+from tqdm.autonotebook import tqdm
 from pathlib import Path
 from cached_property import cached_property
 from functools import partial
@@ -24,6 +24,11 @@ import requests
 import zipfile
 import urllib.parse
 import cachetools
+import sqlite3
+import uuid
+
+
+pbar_on_open = False
 
 
 _gensim_version = int(gensim.__version__.split(".")[0])
@@ -31,7 +36,7 @@ if _gensim_version >= 4:
 	raise RuntimeError("Vectorian needs gensim < 4.0.0")
 
 
-_custom_cache_path = None
+_custom_cache_path = os.environ.get("VECTORIAN_CACHE_HOME")
 
 
 def set_cache_path(path):
@@ -43,11 +48,15 @@ def _make_cache_path():
 	if _custom_cache_path is None:
 		cache_path = Path.home() / ".vectorian" / "embeddings"
 	else:
-		cache_path = _custom_cache_path / "embeddings"
+		cache_path = Path(_custom_cache_path) / "embeddings"
 	cache_path.mkdir(exist_ok=True, parents=True)
 	return cache_path
 
 
+def _extraction_tqdm(tokens, name):
+	return tqdm(tokens, desc=f"Extracting {name}", disable=len(tokens) < 5000)
+
+
 def _download(url, path, force_download=False):
 	path = Path(path)
 	download_path = path / urllib.parse.urlparse(url).path.split("/")[-1]
@@ -90,7 +99,7 @@ def _download(url, path, force_download=False):
 	return result_path if result_path.exists() else None
 
 
-def _normalize_word2vec(tokens, embeddings, normalizer, sampling='nearest'):
+def _normalize_word2vec(name, tokens, embeddings, normalizer, sampling='nearest'):
 	if sampling not in ('nearest', 'average'):
 		raise ValueError(f'Expected "nearest" or "average", got "{sampling}"')
 
@@ -100,7 +109,7 @@ def _normalize_word2vec(tokens, embeddings, normalizer, sampling='nearest'):
 	f_tokens = []
 	token_to_ids = dict()
 
-	for i, t in enumerate(tqdm(tokens, desc="Normalizing Tokens")):
+	for i, t in enumerate(tqdm(tokens, desc=f"Normalizing tokens in {name}")):
 		nt = normalizer(t)
 		if nt is None:
 			continue
@@ -115,7 +124,7 @@ def _normalize_word2vec(tokens, embeddings, normalizer, sampling='nearest'):
 			indices.append(i)
 
 	if sampling == 'average':
-		for indices in tqdm(token_to_ids.values(), desc="Merging Tokens", total=len(token_to_ids)):
+		for indices in tqdm(token_to_ids.values(), desc=f"Merging tokens in {name}", total=len(token_to_ids)):
 			if len(indices) > 1:
 				i = indices[0]
 				embeddings[i] = np.mean(embeddings[indices], axis=0)
@@ -184,6 +193,13 @@ class StaticEmbedding(Embedding):
 
 
 class AbstractVectors:
+	def __enter__(self):
+		return self
+
+	def __exit__(self, exc_type, exc_value, exc_traceback):
+		self.close()
+		return False
+
 	def _save_transform(self, hf):
 		pass
 
@@ -213,13 +229,6 @@ class Vectors(AbstractVectors):
 	def __init__(self, unmodified):
 		self._unmodified = unmodified
 
-	def __enter__(self):
-		return self
-
-	def __exit__(self, exc_type, exc_value, exc_traceback):
-		self.close()
-		return False
-
 	def close(self):
 		pass  # a no op
 
@@ -366,8 +375,6 @@ class StaticEmbeddingInstance:
 class CachedWordEmbedding(StaticEmbedding):
 	class Cache:
 		def __init__(self):
-			import sqlite3
-
 			self._cache_path = _make_cache_path() / 'cache'
 			self._cache_path.mkdir(exist_ok=True, parents=True)
 
@@ -478,7 +485,7 @@ class CachedWordEmbedding(StaticEmbedding):
 			#dat_path = normalized_cache_path / f"{name}-{normalizer.name}-{self._embedding_sampling}.dat"
 
 			if dat_path and dat_path.exists():
-				with tqdm(desc="Opening " + self.name, total=1,  bar_format='{l_bar}{bar}') as pbar:
+				with tqdm(desc="Opening " + self.name, total=1,  bar_format='{l_bar}{bar}', disable=not pbar_on_open) as pbar:
 					with open(dat_path.with_suffix('.json'), 'r') as f:
 						data = json.loads(f.read())
 					tokens = data['tokens']
@@ -488,7 +495,7 @@ class CachedWordEmbedding(StaticEmbedding):
 			else:
 				tokens, vectors = self._load()
 				tokens, vectors = _normalize_word2vec(
-					tokens, vectors, normalizer.unpack(), self._embedding_sampling)
+					self.name, tokens, vectors, normalizer.unpack(), self._embedding_sampling)
 
 				for t in self._transforms:
 					vectors = t.apply(Vectors(vectors)).unmodified
@@ -602,7 +609,7 @@ class KeyedVectors(StaticEmbedding):
 
 		def get_embeddings(self, tokens):
 			data = np.empty((len(tokens), self.dimension), dtype=np.float32)
-			for i, t in tqdm(enumerate(tokens), disable=len(tokens) < 1000):
+			for i, t in enumerate(_extraction_tqdm(tokens, self.name)):
 				data[i, :] = self._wv.word_vec(t)
 			return Vectors(data)
 
@@ -683,8 +690,14 @@ class Zoo:
 				'name': f'numberbatch-19.08-{lang}'
 			}
 
+		for d in [50, 100, 200, 300]:
+			Zoo._embeddings[f'glove-6B-{d}'] = {
+				'constructor': partial(Word2VecVectors, binary=True),
+				'url': _zenodo_url(4925376, f'glove.6B.{d}d.zip'),
+				'name': f'glove-6B-{d}'
+			}
+
 		for name, sizes in {
-			'6B': [50, 100, 200, 300],
 			'42B': [300],
 			'840B': [300],
 			'twitter.27B': [25, 50, 100, 200]}.items():
@@ -696,14 +709,6 @@ class Zoo:
 					'ndims': size
 				}
 
-		# add special glove source for faster downloads.
-
-		Zoo._embeddings['glove-6B-50'] = {
-			'constructor': PretrainedGensimVectors,
-			'name': 'glove-6B-50',
-			'gensim_name': 'glove-wiki-gigaword-50'
-		}
-
 		Zoo._initialized = True
 
 	@staticmethod
@@ -769,7 +774,7 @@ class PretrainedFastText(StaticEmbedding):
 
 		def get_embeddings(self, tokens):
 			data = np.empty((len(tokens), self.dimension), dtype=np.float32)
-			for i, t in tqdm(enumerate(tokens), disable=len(tokens) < 1000):
+			for i, t in enumerate(_extraction_tqdm(tokens, self.name)):
 				data[i, :] = self._ft.get_word_vector(t)
 			return Vectors(data)
 
@@ -858,7 +863,7 @@ class StackedEmbedding(StaticEmbedding):
 
 		def get_embeddings(self, tokens):
 			data = np.empty((len(tokens), self.dimension), dtype=np.float32)
-			for i, t in tqdm(enumerate(tokens), disable=len(tokens) < 1000):
+			for i, t in enumerate(_extraction_tqdm(tokens, self.name)):
 				data[i, :] = self.word_vec(t)
 			return Vectors(data)
 
@@ -961,23 +966,79 @@ class SpacyEmbedding(ContextualEmbedding):
 		] + ([] if self._transform is None else [self._transform.name]))
 
 
+class VectorCache:
+	def __init__(self, path, readonly=False):
+		self._path = Path(path)
+		self._readonly = readonly
+
+		self._conn = sqlite3.connect(self._path / 'cache.db')
+		self._conn.execute("create table if not exists cache (key varchar primary key, stem varchar)")
+
+	def _get_stem(self, key):
+		cur = self._conn.cursor()
+		try:
+			cur.execute('select stem from cache where key=?', (key, ))
+			r = cur.fetchone()
+		finally:
+			cur.close()
+
+		if r is None:
+			return None
+		else:
+			return r[0]
+
+	def put(self, key, array):
+		if self._readonly:
+			return
+
+		stem = self._get_stem(key)
+		if stem is not None:
+			npy_path = self._path / (stem + ".npy")
+			np.save(npy_path, array)
+		else:
+			stem = uuid.uuid1().hex
+			npy_path = self._path / (stem + ".npy")
+			assert not npy_path.exists()
+
+			with self._conn:
+				self._conn.execute("insert into cache(key, stem) values (?, ?)", (
+					key, stem))
+
+				np.save(npy_path, array)
+
+	def get(self, key):
+		stem = self._get_stem(key)
+		if stem is None:
+			return None
+		else:
+			return np.load(self._path / (stem + ".npy"))
+
+
 class SpacyVectorEmbedding(SpacyEmbedding):
-	def __init__(self, nlp, dimension, **kwargs):
+	def __init__(self, nlp, dimension, cache=None, **kwargs):
 		super().__init__(nlp, **kwargs)
 		self._dimension = dimension
+		self._cache = cache
 
 	@property
 	def dimension(self):
 		return self._dimension
 
 	def pca(self, n_dims):
-		return SentenceBertEmbedding(self._nlp, PCACompression(n_dims))
+		return SpacyVectorEmbedding(self._nlp, PCACompression(n_dims))
 
 	def encode(self, doc):
-		return np.array([token.vector for token in doc])
+		if self._cache is not None:
+			array = self._cache.get(doc.text)
+			if array is not None:
+				return array
 
+		array = np.array([token.vector for token in doc])
 
-SentenceBertEmbedding = SpacyVectorEmbedding  # legacy
+		if self._cache is not None:
+			self._cache.put(doc.text, array)
+
+		return array
 
 
 class SpacyTransformerEmbedding(SpacyEmbedding):
@@ -1076,7 +1137,7 @@ class ExternalMemoryVectors:
 		return vectors
 
 
-class VectorsCache:
+class OpenedVectorsCache:
 	def __init__(self):
 		self._cache = dict()
 
@@ -1120,6 +1181,9 @@ class ProxyVectorsRef(VectorsRef):
 	def open(self):
 		return self._vectors
 
+	def to_internal_memory(self):
+		return self
+
 
 class ExternalMemoryVectorsRef(VectorsRef):
 	def __init__(self, path):
@@ -1128,6 +1192,10 @@ class ExternalMemoryVectorsRef(VectorsRef):
 	def open(self):
 		return ExternalMemoryVectors.load(self._path)
 
+	def to_internal_memory(self):
+		with self.open() as vectors:
+			return ProxyVectorsRef(Vectors(vectors.unmodified))
+
 
 class MaskedVectorsRef(VectorsRef):
 	def __init__(self, vectors, mask):
@@ -1138,6 +1206,11 @@ class MaskedVectorsRef(VectorsRef):
 		return MaskedVectors(
 			self._vectors.open(), self._mask)
 
+	def to_internal_memory(self):
+		return MaskedVectorsRef(
+			self._vectors.to_internal_memory(), self._mask)
+
+
 # === sentence/partition encoders.
 
 
